import numpy as np
from .base import IOHandler
from ...core import Molecule

class XyzIO(IOHandler):
    '''The XYZ format is described in this wikipedia article
    http://en.wikipedia.org/wiki/XYZ_file_format.
    
    **Features**

    .. method:: read("molecule")
    
       Read the coordinates in a :py:class:`~chemlab.core.Molecule` instance.
       
    .. method:: write("molecule", mol)

       Writes a :py:class:`~chemlab.core.Molecule` instance in the XYZ format.
    '''
    
    can_read = ['molecule']
    can_write = ['molecule']
    
    def read(self, feature):
        self.check_feature(feature, "read")
        
        lines = self.fd.readlines()
        
        num = int(lines[0])
        title = lines[1]

        if feature == 'title':
            return title
            
        if feature == 'molecule':
            type_array = []
            r_array = []
            for l in lines[2:]:
                type, x, y, z = l.split()
                r_array.append([float(x),float(y),float(z)])
                type_array.append(type)
            
            r_array = np.array(r_array)/10 # To nm
            type_array = np.array(type_array)
            
            return Molecule.from_arrays(r_array=r_array, type_array=type_array)
            
            
    def write(self, feature, mol):
        self.check_feature(feature, "write")
        
        lines = []
        if feature == 'molecule':
            lines.append(str(mol.n_atoms))
            
            lines.append('Generated by chemlab')
            for t, (x, y, z) in zip(mol.type_array, mol.r_array):
                lines.append('    %s       %.6f      %.6f      %.6f' %
                             (t, x*10, y*10, z*10))
            
            self.fd.write('\n'.join(lines))
    